/*
/*   Copyright (c) 2006, David S. Madole
/*   All rights reserved.
/*
/*   For latest sources and information on this software, please
/*   go to http://www.madole.net/pushmail
/*
/*   Questions, comments or suggestions should be directed to
/*   David S. Madole <david@madole.net>.
/*
/*   Redistribution and use in source and binary forms, with or without
/*   modification, are permitted provided that the following conditions
/*   are met:
/*
/*   1. Redistributions of source code must retain the above copyright
/*      notice unmodified, this list of conditions, and the following
/*      disclaimer.
/*
/*   2. Redistributions in binary form must reproduce the above copyright
/*      notice, this list of conditions and the following disclaimer in the
/*      documentation and/or other materials provided with the distribution.
/*
/*   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
/*   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
/*   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
/*   ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
/*   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
/*   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
/*   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
/*   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
/*   LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
/*   OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
/*   SUCH DAMAGE.
/*
/*   Initial public release - 08/29/2006
*/


#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <signal.h>
#include <syslog.h>
#include <errno.h>
#include <fcntl.h>
#include <time.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>


/*  This struct is used to keep track of the replica servers and their
//  corresponding log files and statuses. */

struct serventry {

    char *host;		/* Host name */
    char *sync;		/* Host-specific sync log file */
    char *work;		/* Host-specific working log file */
    int syncok;		/* Sync log file is ready for processing */
    int workok;		/* Working log file is ready for processing */
    int client;		/* Process ID of running client, if any */
    int interval;	/* How often to push mail to this replica */
    int schedule;	/* When to next push mail to this replica */
};


int syslogging = 0;


void
logerror(char *fmt, ...) {

    va_list ap;

    va_start(ap, fmt);

    if (syslogging) vsyslog(LOG_ERR, fmt, ap);
    else vfprintf(stderr, fmt, ap);

    va_end(ap);
}


int
lockfd(int fd, int type) /* F_WRLCK, F_UNLCK */
{
    struct flock lock;
    int res;

    for ( ; ; ) {

        lock.l_type = type;
        lock.l_whence = SEEK_SET;
        lock.l_start = 0;
        lock.l_len = 0;

        res = fcntl(fd, F_SETLKW, &lock);

        if (res != -1) return 0;
        if (errno == EINTR) continue;

        return -1;
    }
}


int
append(char *src, char *dst) {

    char *ptr, buf[4096];
    int sfd, dfd, len, res;

    res = -1;

    sfd = open(src, O_RDONLY);

    if (sfd >= 0) {

	if (lockfd(sfd, F_RDLCK) >= 0) {

	    dfd = open(dst, O_WRONLY | O_APPEND | O_CREAT, 0640);

	    if (dfd >= 0) {

		if (lockfd(dfd, F_WRLCK) >= 0) {

		    for ( ; ; ) {

			res = read(sfd, buf, sizeof(buf));

			if (res == 0) break;

			if (res > 0) {

			    for (ptr = buf, len = res; len > 0; ) {

				res = write(dfd, ptr, len);

				if (res < 0) {

				    if (errno == EINTR) continue;
				    break;
				}
				ptr += res;
				len -= res;
			    }
			}

			if (res < 0) {

			    if (errno == EINTR) continue;
			    break;
			}
		    }
		    lockfd(dfd, F_UNLCK);
		}
		close(dfd);
	    }
	    lockfd(sfd, F_UNLCK);
	}
	close(sfd);
    }
    return res;
}


int
startsync(struct serventry *server) {

    int pid, fds[2];
    char buf[1];

    if (pipe(fds) < 0) {

	logerror("pipe() returned %d (%s)\n", errno, strerror(errno));
	return -1;
    }

    switch (pid = fork()) {

    case -1:

	logerror("fork() returned %d (%s)\n", errno, strerror(errno));
	return -1;

    case 0:

	close(fds[1]);
	read(fds[0], buf, 1);
	close(fds[0]);

	execlp("sync_client",
	    "sync_client", "-f", server->work, "-S", server->host, "-r", NULL);

	logerror("execlp() returned %d (%s)\n", errno, strerror(errno));
	return -1;

    default:

	server->client = pid;

	close(fds[0]);
	write(fds[1], buf, 1);
	close(fds[1]);

	return 0;
    }
}


struct serventry *servers;


void sigchild(int sig) {

    struct serventry *server;
    int pid, status;

    for ( ; ; ) {

	pid = wait(&status);

	if (pid != -1) break;

	if (errno != EINTR) {

	    logerror("wait() returned %d (%s)\n", errno, strerror(errno));
	    return;
	}
    }

    for (server = servers; server->host != NULL; server++) {

	if (server->client == pid) {

	    if (WIFEXITED(status) && WEXITSTATUS(status) == EXIT_SUCCESS) {

		if (unlink(server->work) < 0) {

		    logerror("unlink(\"%s\") returned %d (%s)\n",
			server->work, errno, strerror(errno));
		}
		else server->workok = 0;
	    }
	    server->client = 0;
	}
    }
}


int
main(int argc, char **argv) {

    struct serventry *server, master[1];
    struct stat sb;
    int res, fd;
    char *ptr, buf[99];
    time_t now;

    char *synclog = "/var/imap/sync/log";

    char *pidfile = NULL;
    int background = 0;
    int interval = 60;


    /*  We use a replica server data structure to keep track of information
    //  on the master sync logs as well, since the data is so similar. */

    asprintf(&master->sync, "%s", synclog);
    asprintf(&master->work, "%s.work", synclog);

    if (master->sync == NULL || master->work == NULL) {

	logerror("asprintf() returned %d (%s)\n",
	    errno, strerror(errno));
	return EXIT_FAILURE;
    }
    master->interval = 5;
    master->schedule = 0;


    /*  Check for the state of any pre-existing sync log files or working
    //  log files so that we can resume where a previous run might have
    //  left off, or pickup events that occurred before we started. */

    master->syncok = stat(master->sync, &sb) == 0;
    master->workok = stat(master->work, &sb) == 0;


    /*  Allocate an array of replica server data structures. Just use the
    //  number of command-line arguments as an upper-limit, since there can
    //  never be more replicas than arguments. */

    server = servers = malloc(sizeof(*servers) * argc);

    if (servers == NULL) {

	logerror("malloc() returned %d (%s)\n", errno, strerror(errno));
	return EXIT_FAILURE;
    }


    /*  Process command-line arguments, setting options and adding replica
    //  server records as needed. */

    for (argv++; *argv != NULL; argv++) {

	if (strcmp(*argv, "-b") == 0) background = 1;

	else if (strcmp(*argv, "-p") == 0) pidfile = *++argv;

	else if (strcmp(*argv, "-i") == 0) master->interval = atoi(*++argv);

	else if (strcmp(*argv, "-t") == 0) interval = atoi(*++argv);

	else if (**argv == '-') {

	    logerror("unknown option \"%s\" specified\n", *argv);
	    return EXIT_FAILURE;
	}
	else {

	    asprintf(&server->host, "%s", *argv);
	    asprintf(&server->sync, "%s.%s", synclog, *argv);
	    asprintf(&server->work, "%s.%s.work", synclog, *argv);

	    if (server->host == NULL ||
		server->sync == NULL || server->work == NULL) {

		logerror("asprintf() returned %d (%s)\n",
		    errno, strerror(errno));
		return EXIT_FAILURE;
	    }

	    /*  As with the master log files, check for any pre-existing
	    //  replica log files so we can pickup where we left off. */

	    server->syncok = stat(server->sync, &sb) == 0;
	    server->workok = stat(server->work, &sb) == 0;

	    server->interval = interval;
	    server->schedule = 0;

	    server->client = 0;

	    server++;
	}
    }
    server->host = NULL;


    /*  Open the pidfile before daemonizing so that we can report any error
    //  message to the startup environment and return a failure status. */

    if (pidfile) {

	fd = open(pidfile, O_WRONLY | O_CREAT | O_TRUNC, 0640);

	if (fd < 0) {

	    logerror("open(\"%s\") returned %d {%s}\n",
		pidfile, errno, strerror(errno));
	    return EXIT_FAILURE;
	}
    }


    /*  If we are to run in the background, daemonize now by double-forking
    //  to orphan the background process so that it will become a child
    //  process of init. Change to the root directory first so we don't hold
    //  any filesystems open unnecessarily. */

    if (background) {

	if (chdir("/") < 0) {

	    logerror("chdir(\"/\") returned %d {%s}\n",
		errno, strerror(errno));
	    return EXIT_FAILURE;
	}

	res = fork();

	if (res != 0) {

	    if (pidfile) close(fd);

	    if (res < 0) {

		logerror("fork() returned %d {%s}\n", errno, strerror(errno));
		return EXIT_FAILURE;
	    }
	    return EXIT_SUCCESS;
	}

	/*  At this point, we are a child of the original process and can no
	//  longer return a failure status, so for any failure cases, log
	//  a message to syslog and give up. Close the original stdio fds. */

	openlog("pushmail", LOG_PID, LOG_LOCAL6);

	syslogging = 1;

	close(0);
	close(1);
	close(2);

	res = fork();

	if (res != 0) {

	    if (pidfile) close(fd);

	    if (res < 0) {

		logerror("fork() returned %d {%s}\n", errno, strerror(errno));
		return EXIT_FAILURE;
	    }
	    return EXIT_SUCCESS;
	}
    }


    /*  Now that we are daemonized, if necessary, go ahead and write the pid
    //  to the already-opened pidfile if requested. */

    if (pidfile) {

	res = sprintf(buf, "%d\n", getpid());
	res = write(fd, buf, res);

	close(fd);

	if (res < 0) {

	    logerror("write(\"%s\") returned %d (%s)\n",
		pidfile, errno, strerror(errno));
	    return EXIT_FAILURE;
	}
    }


    /*  Install SIGCHILD handler so we know when background sync_client
    //  processes exit. We need to keep track so we don't run more than one
    //  against the same replica at a time. */

    signal(SIGCHLD, sigchild);


    /*  Begin main processing loop, checking sync log and firing off
    //  sync_clients against copies of the log as needed. */

    for( ; ; ) {

	for (server = servers; server->host != NULL; server++) {

	    if (master->workok) {

		if (append(master->work, server->sync) < 0) {

		    logerror("append(\"%s\", \"%s\") returned %d (%s)\n",
			master->work, server->sync, errno, strerror(errno));
		    return EXIT_FAILURE;
		}
		else server->syncok = 1;
	    }

	    if (server->client == 0) {

		if (server->workok) {

		    server->schedule = now + server->interval;
		    startsync(server);
		}
        	else if (server->syncok && server->schedule <= now) {

		    if (rename(server->sync, server->work) < 0) {

			logerror( "rename(\"%s\", \"%s\") returned %d (%s)\n",
			    server->sync, server->work, errno, strerror(errno));
			    return EXIT_FAILURE;
		    }
		    else {

			server->syncok = 0;
			server->workok = 1;

			server->schedule = now + server->interval;
			startsync(server);
		    }
		}
	    }
	}

	if (master->workok) {

            if (unlink(master->work) < 0) {

		logerror("unlink(\"%s\") returned %d (%s)\n",
		    master->work, errno, strerror(errno));
	 	return EXIT_FAILURE;
	    }
	    else master->workok = 0;
	}

	if (master->syncok) {

            if (rename(master->sync, master->work) < 0) {

		logerror("rename(\"%s\", \"%s\") returned %d (%s)\n",
		    master->sync, master->work, errno, strerror(errno));
		return EXIT_FAILURE;
	    }
	    else {

		master->syncok = 0;
		master->workok = 1;
	    }
	}
	else {

	    if (master->schedule <= now) {

		master->syncok = stat(master->sync, &sb) == 0;
		master->schedule = now + master->interval;
	    }
	    else {

		sleep(1);
		now = time(NULL);
	    }
	}
    }
}

